@@ -15,13 +15,23 @@
 
 #include "config.h"
 #include <time.h>
+#include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <limits.h> /* for PATH_MAX */
-#ifdef HAVE_ATTR_XATTR_H
+#include <dirent.h> /* for scandir() and alphasort() */
+#if defined HAVE_SYS_XATTR_H
+#include <sys/xattr.h>
+#elif defined HAVE_ATTR_XATTR_H
 #include <attr/xattr.h>
 #endif
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif
+#ifdef HAVE_SYS_SYSMACROS_H
+#include <sys/sysmacros.h>
+#endif
+
 #include <ext2fs/ext2fs.h>
 #include <ext2fs/ext2_types.h>
 #include <ext2fs/fiemap.h>
@@ -29,7 +39,7 @@
 #include "create_inode.h"
 #include "support/nls-enable.h"
 
-/* 64KiB is the minimium blksize to best minimize system call overhead. */
+/* 64KiB is the minimum blksize to best minimize system call overhead. */
 #define COPY_FILE_BUFLEN	65536
 
 static int ext2_file_type(unsigned int mode)
@@ -111,8 +121,10 @@ static errcode_t set_inode_extra(ext2_filsys fs, ext2_ino_t ino,
 	}
 
 	inode.i_uid = st->st_uid;
+	ext2fs_set_i_uid_high(inode, st->st_uid >> 16);
 	inode.i_gid = st->st_gid;
-	inode.i_mode |= st->st_mode;
+	ext2fs_set_i_gid_high(inode, st->st_gid >> 16);
+	inode.i_mode = (LINUX_S_IFMT & inode.i_mode) | (~S_IFMT & st->st_mode);
 	inode.i_atime = st->st_atime;
 	inode.i_mtime = st->st_mtime;
 	inode.i_ctime = st->st_ctime;
@@ -133,6 +145,9 @@ static errcode_t set_inode_xattr(ext2_filsys fs, ext2_ino_t ino,
 	char				*list = NULL;
 	int				i;
 
+	if (no_copy_xattrs)
+		return 0;
+
 	size = llistxattr(filename, NULL, 0);
 	if (size == -1) {
 		retval = errno;
@@ -151,6 +166,13 @@ static errcode_t set_inode_xattr(ext2_filsys fs, ext2_ino_t ino,
 		return retval;
 	}
 
+	retval = ext2fs_xattrs_read(handle);
+	if (retval) {
+		com_err(__func__, retval,
+			_("while reading xattrs for inode %u"), ino);
+		return retval;
+	}
+
 	retval = ext2fs_get_mem(size, &list);
 	if (retval) {
 		com_err(__func__, retval, _("while allocating memory"));
@@ -223,9 +245,10 @@ static errcode_t set_inode_xattr(ext2_filsys fs EXT2FS_ATTR((unused)),
 }
 #endif  /* HAVE_LLISTXATTR */
 
+#ifndef _WIN32
 /* Make a special files (block and character devices), fifo's, and sockets  */
 errcode_t do_mknod_internal(ext2_filsys fs, ext2_ino_t cwd, const char *name,
-			    struct stat *st)
+			    unsigned int st_mode, unsigned int st_rdev)
 {
 	ext2_ino_t		ino;
 	errcode_t		retval;
@@ -233,7 +256,7 @@ errcode_t do_mknod_internal(ext2_filsys fs, ext2_ino_t cwd, const char *name,
 	unsigned long		devmajor, devminor, mode;
 	int			filetype;
 
-	switch(st->st_mode & S_IFMT) {
+	switch(st_mode & S_IFMT) {
 	case S_IFCHR:
 		mode = LINUX_S_IFCHR;
 		filetype = EXT2_FT_CHRDEV;
@@ -246,10 +269,12 @@ errcode_t do_mknod_internal(ext2_filsys fs, ext2_ino_t cwd, const char *name,
 		mode = LINUX_S_IFIFO;
 		filetype = EXT2_FT_FIFO;
 		break;
+#ifndef _WIN32
 	case S_IFSOCK:
 		mode = LINUX_S_IFSOCK;
 		filetype = EXT2_FT_SOCK;
 		break;
+#endif
 	default:
 		return EXT2_ET_INVALID_ARGUMENT;
 	}
@@ -287,8 +312,8 @@ errcode_t do_mknod_internal(ext2_filsys fs, ext2_ino_t cwd, const char *name,
 		fs->now ? fs->now : time(0);
 
 	if (filetype != S_IFIFO) {
-		devmajor = major(st->st_rdev);
-		devminor = minor(st->st_rdev);
+		devmajor = major(st_rdev);
+		devminor = minor(st_rdev);
 
 		if ((devmajor < 256) && (devminor < 256)) {
 			inode.i_block[0] = devmajor * 256 + devminor;
@@ -307,6 +332,7 @@ errcode_t do_mknod_internal(ext2_filsys fs, ext2_ino_t cwd, const char *name,
 
 	return retval;
 }
+#endif
 
 /* Make a symlink name -> target */
 errcode_t do_symlink_internal(ext2_filsys fs, ext2_ino_t cwd, const char *name,
@@ -392,7 +418,7 @@ static ssize_t my_pread(int fd, void *buf, size_t count, off_t offset)
 }
 #endif /* !defined HAVE_PREAD64 && !defined HAVE_PREAD */
 
-static errcode_t copy_file_range(ext2_filsys fs, int fd, ext2_file_t e2_file,
+static errcode_t copy_file_chunk(ext2_filsys fs, int fd, ext2_file_t e2_file,
 				 off_t start, off_t end, char *buf,
 				 char *zerobuf)
 {
@@ -422,8 +448,8 @@ static errcode_t copy_file_range(ext2_filsys fs, int fd, ext2_file_t e2_file,
 				ptr += blen;
 				continue;
 			}
-			err = ext2fs_file_lseek(e2_file, off + bpos,
-						EXT2_SEEK_SET, NULL);
+			err = ext2fs_file_llseek(e2_file, off + bpos,
+						 EXT2_SEEK_SET, NULL);
 			if (err)
 				goto fail;
 			while (blen > 0) {
@@ -464,9 +490,10 @@ static errcode_t try_lseek_copy(ext2_filsys fs, int fd, struct stat *statbuf,
 		if (hole < 0)
 			return EXT2_ET_UNIMPLEMENTED;
 
-		data_blk = data & ~(fs->blocksize - 1);
-		hole_blk = (hole + (fs->blocksize - 1)) & ~(fs->blocksize - 1);
-		err = copy_file_range(fs, fd, e2_file, data_blk, hole_blk, buf,
+		data_blk = data & ~(off_t)(fs->blocksize - 1);
+		hole_blk = ((hole + (off_t)(fs->blocksize - 1)) &
+			    ~(off_t)(fs->blocksize - 1));
+		err = copy_file_chunk(fs, fd, e2_file, data_blk, hole_blk, buf,
 				      zerobuf);
 		if (err)
 			return err;
@@ -510,13 +537,14 @@ static errcode_t try_fiemap_copy(ext2_filsys fs, int fd, ext2_file_t e2_file,
 		if (err < 0 && (errno == EOPNOTSUPP || errno == ENOTTY)) {
 			err = EXT2_ET_UNIMPLEMENTED;
 			goto out;
-		} else if (err < 0 || fiemap_buf->fm_mapped_extents == 0) {
+		} else if (err < 0) {
 			err = errno;
 			goto out;
-		}
+		} else if (fiemap_buf->fm_mapped_extents == 0)
+			goto out;
 		for (i = 0, ext = ext_buf; i < fiemap_buf->fm_mapped_extents;
 		     i++, ext++) {
-			err = copy_file_range(fs, fd, e2_file, ext->fe_logical,
+			err = copy_file_chunk(fs, fd, e2_file, ext->fe_logical,
 					      ext->fe_logical + ext->fe_length,
 					      buf, zerobuf);
 			if (err)
@@ -569,7 +597,7 @@ static errcode_t copy_file(ext2_filsys fs, int fd, struct stat *statbuf,
 		goto out;
 #endif
 
-	err = copy_file_range(fs, fd, e2_file, 0, statbuf->st_size, buf,
+	err = copy_file_chunk(fs, fd, e2_file, 0, statbuf->st_size, buf,
 			      zerobuf);
 out:
 	ext2fs_free_mem(&zerobuf);
@@ -641,7 +669,7 @@ errcode_t do_write_internal(ext2_filsys fs, ext2_ino_t cwd, const char *src,
 		com_err(__func__, 0, "Warning: inode already set");
 	ext2fs_inode_alloc_stats2(fs, newfile, +1, 0);
 	memset(&inode, 0, sizeof(inode));
-	inode.i_mode = (statbuf.st_mode & ~LINUX_S_IFMT) | LINUX_S_IFREG;
+	inode.i_mode = (statbuf.st_mode & ~S_IFMT) | LINUX_S_IFREG;
 	inode.i_atime = inode.i_ctime = inode.i_mtime =
 		fs->now ? fs->now : time(0);
 	inode.i_links_count = 1;
@@ -678,14 +706,98 @@ out:
 	return retval;
 }
 
-/* Copy files from source_dir to fs */
+struct file_info {
+	char *path;
+	size_t path_len;
+	size_t path_max_len;
+};
+
+static errcode_t path_append(struct file_info *target, const char *file)
+{
+	if (strlen(file) + target->path_len + 1 > target->path_max_len) {
+		void *p;
+		target->path_max_len *= 2;
+		p = realloc(target->path, target->path_max_len);
+		if (p == NULL)
+			return EXT2_ET_NO_MEMORY;
+		target->path = p;
+	}
+	target->path_len += sprintf(target->path + target->path_len, "/%s",
+				    file);
+	return 0;
+}
+
+#ifdef _WIN32
+static int scandir(const char *dir_name, struct dirent ***name_list,
+		   int (*filter)(const struct dirent*),
+		   int (*compar)(const struct dirent**, const struct dirent**)) {
+	DIR *dir;
+	struct dirent *dent;
+	struct dirent **temp_list = NULL;
+	size_t temp_list_size = 0; // unit: num of dirent
+	size_t num_dent = 0;
+
+	dir = opendir(dir_name);
+	if (dir == NULL) {
+		return -1;
+	}
+
+	while ((dent = readdir(dir))) {
+		if (filter != NULL && !(*filter)(dent))
+			continue;
+
+		// re-allocate the list
+		if (num_dent == temp_list_size) {
+			size_t new_list_size = temp_list_size + 32;
+			struct dirent **new_list = (struct dirent**)realloc(
+				temp_list, new_list_size * sizeof(struct dirent*));
+			if (new_list == NULL) {
+				goto out;
+			}
+			temp_list_size = new_list_size;
+			temp_list = new_list;
+		}
+		// add the copy of dirent to the list
+		temp_list[num_dent] = (struct dirent*)malloc((dent->d_reclen + 3) & ~3);
+		memcpy(temp_list[num_dent], dent, dent->d_reclen);
+		num_dent++;
+	}
+
+	if (compar != NULL) {
+		qsort(temp_list, num_dent, sizeof(struct dirent*),
+		      (int (*)(const void*, const void*))compar);
+	}
+
+        // release the temp list
+	*name_list = temp_list;
+	temp_list = NULL;
+
+out:
+	if (temp_list != NULL) {
+		while (num_dent > 0) {
+			free(temp_list[--num_dent]);
+		}
+		free(temp_list);
+		num_dent = -1;
+	}
+	closedir(dir);
+	return num_dent;
+}
+
+static int alphasort(const struct dirent **a, const struct dirent **b) {
+	return strcoll((*a)->d_name, (*b)->d_name);
+}
+#endif
+
+/* Copy files from source_dir to fs in alphabetical order */
 static errcode_t __populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
 			       const char *source_dir, ext2_ino_t root,
-			       struct hdlinks_s *hdlinks)
+			       struct hdlinks_s *hdlinks,
+			       struct file_info *target,
+			       struct fs_ops_callbacks *fs_callbacks)
 {
 	const char	*name;
-	DIR		*dh;
-	struct dirent	*dent;
+	struct dirent	**dent;
 	struct stat	st;
 	char		*ln_target = NULL;
 	unsigned int	save_inode;
@@ -693,6 +805,8 @@ static errcode_t __populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
 	errcode_t	retval = 0;
 	int		read_cnt;
 	int		hdlink;
+	size_t		cur_dir_path_len;
+	int		i, num_dents;
 
 	if (chdir(source_dir) < 0) {
 		retval = errno;
@@ -702,24 +816,25 @@ static errcode_t __populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
 		return retval;
 	}
 
-	if (!(dh = opendir("."))) {
+	num_dents = scandir(".", &dent, NULL, alphasort);
+
+	if (num_dents < 0) {
 		retval = errno;
 		com_err(__func__, retval,
-			_("while opening directory \"%s\""), source_dir);
+			_("while scanning directory \"%s\""), source_dir);
 		return retval;
 	}
 
-	while ((dent = readdir(dh))) {
-		if ((!strcmp(dent->d_name, ".")) ||
-		    (!strcmp(dent->d_name, "..")))
+	for (i = 0; i < num_dents; free(dent[i]), i++) {
+		name = dent[i]->d_name;
+		if ((!strcmp(name, ".")) || (!strcmp(name, "..")))
 			continue;
-		if (lstat(dent->d_name, &st)) {
+		if (lstat(name, &st)) {
 			retval = errno;
 			com_err(__func__, retval, _("while lstat \"%s\""),
-				dent->d_name);
+				name);
 			goto out;
 		}
-		name = dent->d_name;
 
 		/* Check for hardlinks */
 		save_inode = 0;
@@ -740,12 +855,30 @@ static errcode_t __populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
 				save_inode = 1;
 		}
 
+		cur_dir_path_len = target->path_len;
+		retval = path_append(target, name);
+		if (retval) {
+			com_err(__func__, retval,
+				"while appending %s", name);
+			goto out;
+		}
+
+		if (fs_callbacks && fs_callbacks->create_new_inode) {
+			retval = fs_callbacks->create_new_inode(fs,
+				target->path, name, parent_ino, root,
+				st.st_mode & S_IFMT);
+			if (retval)
+				goto out;
+		}
+
 		switch(st.st_mode & S_IFMT) {
 		case S_IFCHR:
 		case S_IFBLK:
 		case S_IFIFO:
+#ifndef _WIN32
 		case S_IFSOCK:
-			retval = do_mknod_internal(fs, parent_ino, name, &st);
+			retval = do_mknod_internal(fs, parent_ino, name,
+						   st.st_mode, st.st_rdev);
 			if (retval) {
 				com_err(__func__, retval,
 					_("while creating special file "
@@ -788,6 +921,7 @@ static errcode_t __populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
 				goto out;
 			}
 			break;
+#endif
 		case S_IFREG:
 			retval = do_write_internal(fs, parent_ino, name, name,
 						   root);
@@ -817,7 +951,8 @@ find_lnf:
 					goto out;
 			}
 			/* Populate the dir recursively*/
-			retval = __populate_fs(fs, ino, name, root, hdlinks);
+			retval = __populate_fs(fs, ino, name, root, hdlinks,
+					       target, fs_callbacks);
 			if (retval)
 				goto out;
 			if (chdir("..")) {
@@ -853,6 +988,14 @@ find_lnf:
 			goto out;
 		}
 
+		if (fs_callbacks && fs_callbacks->end_create_new_inode) {
+			retval = fs_callbacks->end_create_new_inode(fs,
+				target->path, name, parent_ino, root,
+				st.st_mode & S_IFMT);
+			if (retval)
+				goto out;
+		}
+
 		/* Save the hardlink ino */
 		if (save_inode) {
 			/*
@@ -878,16 +1021,21 @@ find_lnf:
 			hdlinks->hdl[hdlinks->count].dst_ino = ino;
 			hdlinks->count++;
 		}
+		target->path_len = cur_dir_path_len;
+		target->path[target->path_len] = 0;
 	}
 
 out:
-	closedir(dh);
+	for (; i < num_dents; free(dent[i]), i++);
+	free(dent);
 	return retval;
 }
 
-errcode_t populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
-		      const char *source_dir, ext2_ino_t root)
+errcode_t populate_fs2(ext2_filsys fs, ext2_ino_t parent_ino,
+		       const char *source_dir, ext2_ino_t root,
+		       struct fs_ops_callbacks *fs_callbacks)
 {
+	struct file_info file_info;
 	struct hdlinks_s hdlinks;
 	errcode_t retval;
 
@@ -905,8 +1053,28 @@ errcode_t populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
 		return retval;
 	}
 
-	retval = __populate_fs(fs, parent_ino, source_dir, root, &hdlinks);
+	file_info.path_len = 0;
+	file_info.path_max_len = 255;
+	file_info.path = calloc(file_info.path_max_len, 1);
 
+	retval = set_inode_xattr(fs, root, source_dir);
+	if (retval) {
+		com_err(__func__, retval,
+			_("while copying xattrs on root directory"));
+		goto out;
+	}
+
+	retval = __populate_fs(fs, parent_ino, source_dir, root, &hdlinks,
+			       &file_info, fs_callbacks);
+
+out:
+	free(file_info.path);
 	free(hdlinks.hdl);
 	return retval;
 }
+
+errcode_t populate_fs(ext2_filsys fs, ext2_ino_t parent_ino,
+		      const char *source_dir, ext2_ino_t root)
+{
+	return populate_fs2(fs, parent_ino, source_dir, root, NULL);
+}
