@@ -7,5 +7,5 @@
  * file may be redistributed under the GNU Public License v2.
  */
 
-#define E2FSPROGS_VERSION "1.46.0"
-#define E2FSPROGS_DATE "29-Jan-2020"
+#define E2FSPROGS_VERSION "1.46.1"
+#define E2FSPROGS_DATE "9-Feb-2021"
